""" AppOptics APM instrumentation loader

Checks appoptics_apm.config['inst_enabled'] and imports as requested. used by middleware
and djangoware.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
from appoptics_apm import util

def _enabled(m):
    return util.config['inst_enabled'][m]


def load_inst_modules():
    if _enabled('memcache'):
        from appoptics_apm import inst_memcache
    if _enabled('pymongo'):
        from appoptics_apm import inst_pymongo
    if _enabled('sqlalchemy'):
        from appoptics_apm import inst_sqlalchemy
    if _enabled('httplib'):
        from appoptics_apm import inst_httplib
    if _enabled('redis'):
        from appoptics_apm import inst_redis
    # additionally, in djangoware.py: 'django_orm', 'django_templates'



""" AppOptics APM instrumentation for Tornado.

Copyright (C) 2016 by SolarWinds, LLC.
All rights reserved.
"""
# useful methods for instrumenting Tornado
from __future__ import with_statement

import functools
import time

import tornado.web

from appoptics_apm import util
from appoptics_apm import async

appoptics_apm_logger = util.logger

# instrumentation functions for tornado.web.RequestHandler
def request_handler_start(self):
    """ runs from the main HTTP server thread (doesn't set/get Context)

        takes 'self' parameter, which is the current RequestHandler
        instance (which holds the current HTTPRequest in self.request)
    """
    start_time = int(time.time() * 1e6)

    # check for X-Trace header in HTTP request
    ctx, evt = util.Context.start_trace('tornado', xtr=self.request.headers.get("X-Trace"))

    if ctx.is_sampled() and evt.is_valid():
        if hasattr(self, '__class__') and hasattr(self.__class__, '__name__'):
            evt.add_info("Controller", self.__class__.__name__)
            evt.add_info("Action", self.request.method.lower())
        evt.add_info("URL", self.request.uri)
        evt.add_info("Method", self.request.method)
        evt.add_info("HTTP-Host", self.request.host)
        ctx.report(evt)
        ctx.set_as_default()
    self.request._appoptics_apm_ctx = ctx
    # XXX
    self.set_header("X-Trace", str(ctx))

    appoptics_apm_logger.debug('tornado_oboe::request_handler_start trace: {tr}'.format(tr = str(ctx)))
    self.request._appoptics_apm_span_start = start_time


RequestHandler_start = request_handler_start


def request_handler_finish(self):
    """ runs from the main HTTP server thread, or from write/flush() callback
        doesn't set/get Context; just checks if finish event was set by appoptics_apm_start()
    """
    if hasattr(self, 'get_status'):  # recent Tornado
        status_code = self.get_status()
    elif hasattr(self, '_status_code'):  # older Tornado
        status_code = self._status_code
    else:
        status_code = 500

    try:
        ctx = getattr(self.request, '_appoptics_apm_ctx', util.Context.get_default())
        start_time = getattr(self.request, '_appoptics_apm_span_start', None)
        transaction_name = ctx.get_transaction_name()
        if not transaction_name:
            transaction_name = '{controller}.{action}'.format(
                controller=self.__class__.__name__,
                action=self.request.method.lower()
            )
        setattr(util.Context.transaction_dict, 'url_tran', self.request.uri)
        setattr(util.Context.transaction_dict, 'domain', self.request.host)
        setattr(util.Context.transaction_dict, 'transaction_name', transaction_name)
        setattr(util.Context.transaction_dict, 'start_time', start_time)
        setattr(util.Context.transaction_dict, 'request_method', self.request.method)
        setattr(util.Context.transaction_dict, 'status_code', status_code)

        self.set_header("X-Trace", str(ctx))
        util.end_http_trace(ctx.layer)
    except Exception as e:
        appoptics_apm_logger.error("AppOptics APM tornado_oboeerror: %s" % str(e))
    # clear the stored appoptics_apm event/metadata from the request object
    self.request._appoptics_apm_ctx = None

RequestHandler_finish = request_handler_finish


# instrumentation for tornado.httpclient.AsyncHTTPClient
def AsyncHTTPClient_start(request):
    """ takes 'request' param, which is the outgoing HTTPRequest, not the request currently being handled """
    # this is called from AsyncHTTPClient.fetch(), which runs from the RequestHandler's context
    util.log("entry", "cURL", keys={'cURL_URL': request.url, 'Async': True})
    ctx = util.Context.get_default()
    if hasattr(request, 'headers'):
        if hasattr(request.headers, '__setitem__'):  # could be dict or tornado.httputil.HTTPHeaders
            request.headers['X-Trace'] = str(ctx)  # add X-Trace header to outgoing request

    if ctx.is_sampled():
        request._appoptics_apm_ctx = ctx.copy()


def AsyncHTTPClient_finish(request, callback=None, headers=None):
    """
    fires exit event for Async HTTP requests.

    checks for wrapped metadata stored in user's callback function: if
    it exists, that metadata is used & updated when reporting the
    event, so that the callback will "happen after" the exit event.
    """
    if hasattr(callback, '_appoptics_apm_ctx'):  # wrapped callback contains md
        ev = callback._appoptics_apm_ctx.create_event('exit', 'cURL')  # adds edge to md
        if hasattr(request, '_appoptics_apm_ctx'):  # add edge to entry event for this async HTTP call
            ev.add_edge(request._appoptics_apm_ctx)
        mdobj = callback

    elif hasattr(request, '_appoptics_apm_ctx'):  # callback contains no metadata, but request obj does
        ev = request._appoptics_apm_ctx.create_event('exit', 'cURL')
        mdobj = request

    else:  # no metadata found
        return

    if headers and hasattr(headers, 'get') and headers.get('X-Trace', None):
        response_md = headers.get('X-Trace')
        ev.add_edge_str(response_md)  # add response X-Trace header

    mdobj._appoptics_apm_ctx.report(ev)  # increments metadata in mdobj


# used for wrapping stack contexts in Tornado v1.2 stack_context.py
class AppOpticsApmContextWrapper(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped
        # get current context at wrap time (e.g. when preparing "done" callback for an async call)
        if util.Context.get_default().is_sampled():
            # store wrap-time context for use at call time
            self._appoptics_apm_ctx = util.Context.get_default().copy()

    def __call__(self, *args, **kwargs):
        with async.AppOpticsApmContextManager(self):  # uses self._appoptics_apm_ctx as context
            return self.wrapped.__call__(*args, **kwargs)


# replacement for _StackContextWrapper in Tornado v2.x stack_context.py
class _StackContextWrapper(functools.partial):
    def __init__(self, *args, **kwargs):
        if util.Context.get_default().is_sampled():
            self._appoptics_apm_ctx = util.Context.get_default().copy()
        super(_StackContextWrapper, self).__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with async.AppOpticsApmContextManager(self):
            return super(_StackContextWrapper, self).__call__(*args, **kwargs)


class AppOpticsApmBaseHandler(tornado.web.RequestHandler):
    """The base request handler to insert instrumentation code to the normal request 
    handling processes. All the request handlers of the application need to inherit 
    from this base class.
    """

    def prepare(self):
        """This function will be called by tornado.web.Application before the user 
        defined handler. If a handler inherits from this base handler, it should call
        this function in its own prepare() function."""
        super(AppOpticsApmBaseHandler, self).prepare()
        request_handler_start(self)

    def on_finish(self):
        """This function will be called by tornado.web.Application after the user defined
        handler. If a handler inherits from this base handler, it should call this function
        in its own prepare() function."""
        super(AppOpticsApmBaseHandler, self).on_finish()
        request_handler_finish(self)


util.report_layer_init('tornado')



#!/usr/bin/env python
"""
 Copyright (C) 2016 by SolarWinds, LLC.
 All rights reserved.
"""

import sys
import os
from distutils import log
from distutils.command.build import build
from setuptools import setup, Extension
from setuptools.command.install import install
from setuptools.command.sdist import sdist
from setuptools.command.build_ext import build_ext

version = '3.4.2'

def is_alpine_distro():
    # checking alpine dist
    if os.path.exists("/etc/alpine-release"):
        return True

    try:
        with open("/etc/os-release", 'r') as osr:
            releases = osr.readlines()
            releases = [ l[:-1] for l in releases ]
        if 'NAME="Alpine Linux"' in releases:
            return True
    except Exception:
        pass

    return False

is_alpine = is_alpine_distro()
is_darwin = sys.platform == 'darwin'

def python_version_supported():
    if sys.version_info[0] == 2:
        if sys.version_info[1] not in (7,):
            return False
    elif sys.version_info[0] == 3:
        if sys.version_info[1] < 4:
            return False
    return True

def choose_oboe():
    log.info("[SETUP] choose to link to platform specific liboboe")

    liboboe = "liboboe-1.0-x86_64.so.0.0.0"
    alpineoboe = "liboboe-1.0-alpine-x86_64.so.0.0.0"
    tmp_oboe= "/".join(["./appoptics_apm/swig",  liboboe])
    tmp_alpine = "/".join(["./appoptics_apm/swig",  alpineoboe])

    # choose platform compatible oboe and rename it to default name
    try:
        if is_alpine:
            if os.path.exists(tmp_oboe):
                os.remove(tmp_oboe)
            if os.path.exists(tmp_alpine):
                os.renames(tmp_alpine, tmp_oboe)
        elif is_darwin:
            if os.path.exists(tmp_oboe):
                os.remove(tmp_oboe)
            if os.path.exists(tmp_alpine):
                os.remove(tmp_alpine)
        else:
            # assume all others are g-libc compatible
            if os.path.exists(tmp_alpine):
                os.remove(tmp_alpine)

        # liboboe should have been removed for darwin, create symlink for others
        if os.path.exists(tmp_oboe):
            os.symlink(liboboe, "./appoptics_apm/swig/liboboe-1.0.so")
            os.symlink(liboboe, "./appoptics_apm/swig/liboboe-1.0.so.0")
    except Exception  as ecp:
        log.info("[SETUP] failed to set up platform specific oboe... {e}".format(e=ecp))

def os_supported():
    return sys.platform.startswith('linux')


if not (python_version_supported() and os_supported()):
    log.warn("[SETUP] This package supports only Python 2.7, 3.3 and above on Linux"
             "Other platform or python versions may not work as expected.")

ext_modules = [Extension('appoptics_apm.swig._oboe',
                         sources=['appoptics_apm/swig/oboe_wrap.cxx'],
                         depends=['appoptics_apm/swig/oboe.hpp'],
                         include_dirs=['appoptics_apm/swig', 'appoptics_apm'],
                         libraries=['oboe-1.0', 'rt'],
                         library_dirs=['appoptics_apm/swig'],
                         runtime_library_dirs=['$ORIGIN']), ]

class CustomBuild(build):
    def run(self):
        choose_oboe()
        self.run_command('build_ext')
        build.run(self)

class CustomBuildExt(build_ext):
    def run(self):
        if is_darwin:
            return
        build_ext.run(self)

class CustomInstall(install):
    def run(self):
        install.run(self)

class CustomSdist(sdist):
    def run(self):
        log.info("remove liboboe links...")
        os.remove("./appoptics_apm/swig/liboboe-1.0.so")
        os.remove("./appoptics_apm/swig/liboboe-1.0.so.0")
        sdist.run(self)

try:
    import pypandoc
    rst_long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    # fail "gracefully"
    rst_long_description = open('README.md').read()

setup(name='appoptics_apm',
      cmdclass={'build': CustomBuild, 'build_ext': CustomBuildExt, 'install': CustomInstall, 'sdist': CustomSdist},
      version=version,
      author='SolarWinds, LLC',
      author_email='support@appoptics.com',
      url='https://www.appoptics.com/monitor/python-performance',
      download_url='https://pypi.python.org/pypi/appoptics_apm',
      description='AppOptics APM libraries, instrumentation, and web middleware components '
                  'for WSGI, Django, and Tornado.',
      long_description=rst_long_description,
      keywords='appoptics_apm traceview tracelytics oboe liboboe instrumentation performance wsgi middleware django',
      ext_modules=ext_modules,
      packages=['appoptics_apm', 'appoptics_apm.swig'],
      package_data={'appoptics_apm': ['swig/liboboe-1.0.so.0', 'swig/VERSION']},
      license='LICENSE.txt',
      install_requires=['decorator', 'six'])



""" AppOptics APM instrumentation for SQLAlchemy.

 Copyright (C) 2016 by SolarWinds, LLC.
 All rights reserved.
"""
from appoptics_apm import util

appoptics_apm_logger = util.logger


def main():
    dialect_wrappers = {
        'do_commit': do_commit_cb,
        'do_rollback': do_rollback_cb
    }

    module_class_mappings = (
        ('sqlalchemy.engine.default', 'DefaultDialect', {
            'do_commit': do_commit_cb,
            'do_execute': do_execute_cb,
            'do_executemany': do_execute_cb,
            'do_rollback': do_rollback_cb
        },),
        ('sqlalchemy.dialects.mysql.base', 'MySQLDialect', dialect_wrappers,),
        ('sqlalchemy.dialects.postgresql.base', 'PGDialect', dialect_wrappers,),
    )

    for mod, classname, mappings in module_class_mappings:
        try:
            cls = getattr(__import__(mod, fromlist=[classname]), classname)
        except (AttributeError, ImportError):
            pass
        else:
            wrap_methods(cls, mappings)


def wrap_methods(cls, mappings):
    for name, fn in mappings.items():
        try:
            base_method = getattr(cls, name)
        except AttributeError:
            appoptics_apm_logger.error('Failed to patch %s on %s', name, cls.__name__)
        else:
            appoptics_apm_fn = util.log_method(
                'sqlalchemy',
                store_backtrace=util._collect_backtraces('sqlalchemy'),
                callback=fn)(base_method)
            setattr(cls, name, appoptics_apm_fn)


def do_commit_cb(_f, args, _kwargs, _ret):
    self, conn = args[:2]
    return base_info(self.name, conn, 'COMMIT')


def do_execute_cb(_f, args, _kwargs, _ret):
    self, cursor, stmt, params = args[:4]
    info = base_info(self.name, cursor.connection, stmt)
    if not util.config.get('enable_sanitize_sql', True):
        info['QueryArgs'] = str(params)
    return info


def do_rollback_cb(_f, args, _kwargs, _ret):
    self, conn = args[:2]
    return base_info(self.name, conn, 'ROLLBACK')


def base_info(flavor_name, conn, query):
    # This could be a real connection object, or a connection fairy (proxy)
    conn = getattr(conn, 'connection', conn)
    info = {
        'Flavor': flavor_name,
        'Query': query,
    }
    try:
        info['RemoteHost'] = remotehost_from_connection(flavor_name, conn)
    except Exception:
        pass
    return info


def remotehost_from_connection(flavor_name, conn):
    if flavor_name == 'mysql':
        host_info = conn.get_host_info()

        # "127.0.0.1 via TCP/IP" (MySQLdb)
        if host_info.endswith('TCP/IP'):
            return host_info.split()[0].lower()

        # "socket 127.0.0.1:3306" (pymysql)
        if host_info.startswith('socket'):
            return host_info.split()[1].split(':')[0]

    if flavor_name == 'postgresql':
        host_part = [part for part in conn.dsn.split()
                     if part.startswith('host=')]
        if host_part:
            return host_part[0].split('=')[1]


if util.ready():
    main()
